rm(list=ls())

## HELPER CODE
library(mvtnorm)
L2 = function(arg_theta) sqrt(sum(arg_theta^2))
mse_stats = function(theta, mse, Data, verbose=FALSE) {
  theta_star = Data$theta_star; best_mse = Data$best_mse
  
  mse = log(mse)  # log MSE.
  
  m = length(mse)
  Yt = tail(mse, m-1)
  Yt_1 = head(mse, m-1)
  fit = lm(Yt ~ Yt_1)
  mu = coef(fit)[1]/(1-coef(fit)[2])  # mean-level of MSE
  
  if(verbose) {
    print(confint(fit))
    par(mfrow=c(2, 1))
    plot(theta_star, theta, pch=20)
    lines(theta_star, theta_star, type="l", lty=3, col="gray")
    # plot(theta_star, theta, pch=20)
    plot(mse, type="l", ylim=c(best_mse, max(mse)))
    abline(h=best_mse, col="red", lwd=2, lty=3)
    abline(h=mu, col="blue", lwd=2, lty=3)
    print(sprintf("Best MSE=%.2f -- mean-level MSE = %.2f", best_mse, tail(mse, 1)))
    print("True theta* / Last theta = ")
    print(round(theta_star, 4))
    print(round(theta, 4))
    return(theta)
  } else {
    return(data.frame(mean_level=mu, 
                      max=max(mse),
                      avg=mean(mse),
                      last=tail(mse, 1)))
  }
}


results_line = function(type="normal") {
  par(mfrow=c(1, 1))
  fname = sprintf("experiments/%s_experiment_1.rda", type)
  load(fname)
  require(ggplot2)
  library(plyr)
  library(dplyr)
  # R = subset(results, valtype=="mean_level")
  results = ddply(results, .(gamma1, methods, valtype), summarize, val=mean(val))
  # results$gamma1 = as.factor(results$gamma1)
  results = subset(results, valtype=="mean_level")
  head(results)
  g = ggplot(data=results, aes(x=gamma1, y=val, col=factor(methods)))
  # g = g + geom_line((aes(factor(gamma1))))
  g = g + geom_line()
  #  g = g + scale_x_discrete(labels=round(unique(results$gamma1), 1))
  plot(g)
}


results_boxplot = function(model="normal", whatval="mean_level", 
                           rounding=1, suffix="", title="", only_RM=F) {
  par(mfrow=c(1, 1))
  fname = sprintf("%s_experiment%s.rda", model, suffix)
  load(fname)
  require(ggplot2)
  require(latex2exp)
  head(results)
  # results$gamma1 = as.factor(results$gamma1)
  results = subset(results, valtype==whatval)
  if(only_RM) {
    results = subset(results, methods %in% c("isgd", "sfp"))
  }
  el = element_text(size=18)
  el2 = element_text(size=22)
  el3 = element_text(size=24)
  
  g = ggplot(data=results, aes(x=gamma1, y=val, fill=factor(methods)))
  if(suffix == "_1") {
    g = g + ylim(c(-5, 5))
    el = element_text(size=14)
  }
  
  g = g + geom_boxplot(aes(factor(gamma1))) 
  #g = g+ stat_summary(fun.y=median, geom="line")
  g = g + scale_x_discrete(labels=round(unique(results$gamma1), rounding))
  g = g + labs(fill='method') 
  g= g +  theme(title=el3,
                legend.text=el2,
                legend.title = el2,
                axis.title.x=el,
                axis.text.x =el,
                axis.text.y = el,
                axis.title.y = el)
  g = g + ggtitle(title) + theme(plot.title = element_text(hjust = 0.5))
  
  g = g + xlab(TeX("learning rate $(\\gamma_1)$")) + ylab("log(MSE)")
  
  return(g)
}


all_results_boxplot = function(whatval="mean_level", title="", only_RM=F) {
  par(mfrow=c(1, 1))
  rounding = 1
  fname = "Results_all.rda"
  load(fname)
  results = R0
  results = results[order(results$gamma1), ]
  # results$gamma1 = round(results$gamma1, 3)
  require(ggplot2)
  require(latex2exp)
  head(results)
  if(only_RM) {
    results = subset(results, methods %in% c("isgd", "sfp"))
  }
  # results$gamma1 = as.factor(results$gamma1)
  results = subset(results, valtype==whatval)
  
  el = element_text(size=18)
  el2 = element_text(size=22)
  el3 = element_text(size=24)
  
  g = ggplot(data=results, aes(x=gamma1, y=val, fill=factor(methods)))
  
  
  g = g + geom_boxplot(aes(factor(gamma1))) 
  #g = g+ stat_summary(fun.y=median, geom="line")
  
  g = g + scale_x_discrete(labels=round(unique(results$gamma1), rounding))
  g = g + labs(fill='method') 
  g= g +  theme(title=el3,
                legend.text=el2,
                legend.title = el2,
                axis.title.x=el,
                axis.text.x =el,
                axis.text.y = el,
                axis.title.y = el)
  g = g + ggtitle(title) + theme(plot.title = element_text(hjust = 0.5))
  
  g = g + xlab(TeX("learning rate $(\\gamma_1)$")) + ylab("log(MSE)")
  
  return(g)
}

multiplot <- function(..., plotlist=NULL, file, cols=1, layout=NULL) {
  library(grid)
  
  # Make a list from the ... arguments and plotlist
  plots <- c(list(...), plotlist)
  
  numPlots = length(plots)
  
  # If layout is NULL, then use 'cols' to determine layout
  if (is.null(layout)) {
    # Make the panel
    # ncol: Number of columns of plots
    # nrow: Number of rows needed, calculated from # of cols
    layout <- matrix(seq(1, cols * ceiling(numPlots/cols)),
                     ncol = cols, nrow = ceiling(numPlots/cols))
  }
  
  if (numPlots==1) {
    print(plots[[1]])
    
  } else {
    # Set up the page
    grid.newpage()
    pushViewport(viewport(layout = grid.layout(nrow(layout), ncol(layout))))
    
    # Make each plot, in the correct location
    for (i in 1:numPlots) {
      # Get the i,j matrix positions of the regions that contain this subplot
      matchidx <- as.data.frame(which(layout == i, arr.ind = TRUE))
      
      print(plots[[i]], vp = viewport(layout.pos.row = matchidx$row,
                                      layout.pos.col = matchidx$col))
    }
  }
}
## END OF HELPER CODE.


par(mfrow=c(1, 1))
set.seed(123)
p = 6
n = 10000
kN = 1e8  # Fixed number of iterations for every method.
BUDGET = 1 # every method runs for BUDGET secs.
u = runif(p); 
Sigma = 2 * diag(p) + u %*% t(u)

#' Generates the dataset.
#' @param model Either normal or poisson.
#' 
gen_data = function(model="normal") {
  # θ* and X.
  # Sigma = diag(p) + u %*% t(u)
  all_index = 1:p
  u = runif(p); 
  X = cbind(rmvnorm(n, mean=rep(0, p), sigma = Sigma))
  theta_star = c( exp(-seq(1,p))) * (-1)^all_index
  
  # E(Y|X), regression function for model.
  mean_fn = function(u) u
  if(model=="poisson") {
    mean_fn = function(u) exp(u)
    theta_star = c(exp(-seq(1,p))) # 
  }
  
  X = NA; Y = NA;
  X  =  rmvnorm(n, mean=rep(0, p), sigma = Sigma)
  # Generate data.
  ypred = mean_fn(X %*% theta_star)
  Y  =  rnorm(n, mean=ypred, sd=2) # for normal model
  # Poisson model
  if(model=="poisson") {
    X  =  matrix(sample(c(0, 1, 2, 3), size=n*p, prob=c(0.4, 0.4, 0.15, 0.05), replace=T), nrow=n, ncol=p)
    ypred = mean_fn(X %*% theta_star)
    Y  =  rpois(n, lambda = ypred)
  }
  
  # MLE fit.
  mle_fit = NA
  if(model=="normal") {
    mle_fit = lm(Y ~ X + 0)
  } else if(model=="poisson") {
    mle_fit = glm(Y ~ X + 0, family = "poisson")
  } else {
    stop("Do not have fit function for specified model")
  }
  best_mse = log(L2(coef(mle_fit) - theta_star)) #  MSE of MLE.
  
  list(X=X, Y=Y, theta_star=theta_star,
       mean_fn=mean_fn, mle_fit=mle_fit, best_mse=best_mse)
}

#' H is uniased estimator of the regression function h() as defined in the paper.
#' @param D is the dataset.
#' 
H = function(arg_theta, D) {
  j = sample(1:n, size=1)
  with(D, -(Y[j] - mean_fn(X[j,]*arg_theta)) * X[j,])
}

#' CLassical SGD
#' @param gamma1 The learning rate.
#' @param Data The dataset (Sampled from gen_data)
sgd = function(Data, gamma1=.5, verbose=TRUE) {
  Y = Data$Y; X = Data$X; mean_fn = Data$mean_fn
  theta_star = Data$theta_star
  
  t0 = proc.time()[3]
  theta = rep(0, p)
  mse = c(L2(theta - theta_star))
  for(i in 1:kN) {
    gamma_i = gamma1 / i
    theta = theta - gamma_i * H(theta, Data)
    # Store statistics
    mse = c(mse, L2(theta-theta_star))
    t1 = proc.time()[3]
    if(t1 - t0 > BUDGET) {
      ## over budget
      if(verbose) print(sprintf("Stopped at i=%d / %d. Time = %.2f.  %.3f msec/iter.", 
                                i, kN, t1-t0, 1000*(t1-t0)/ i))
      break;
    }
  }
  
  out = mse_stats(theta, mse, Data, verbose=verbose)  
  return(out)
}

#' Solves the fixed point for the implicit SGD update.
#' 
ISGD_update = function(arg_theta, arg_lr, Data) {
  X = Data$X; Y = Data$Y
  mean_fn = Data$mean_fn
  
  j = sample(1:n, size=1)
  x = X[j,]
  y = Y[j]
  f = function(v) arg_lr * (y - mean_fn(sum(x * arg_theta) + v * sum(x^2)))
  bi = f(0)
  lam = NA
  if(bi == 0) {
    lam = 0
  } else if(bi < 0) { 
    lam = uniroot(function(u) u - f(u), interval = c(bi, 0))$root  
  } else {
    lam = uniroot(function(u) u - f(u), interval = c(0, bi))$root  
  }
  arg_theta + lam * x
}

#' CLassical SGD
#' @param gamma1 The learning rate.
#' @param Data The dataset (Sampled from gen_data)
isgd = function(Data, gamma1=1, verbose=TRUE) {
  theta_star = Data$theta_star
  
  t0 = proc.time()[3]
  theta = rep(0, p)
  mse = c(L2(theta - theta_star))
  for(i in 1:kN) {
    gamma_i = gamma1 / i
    theta = ISGD_update(theta, gamma_i, Data)
    # Statistics.
    mse = c(mse, L2(theta-theta_star))
    t1 = proc.time()[3]
    if(t1 - t0 > BUDGET) {
      ## over budget
      if(verbose) print(sprintf("Stopped at i=%d / %d. Time = %.2f.  %.3f msec/iter.", 
                                i, kN, t1-t0, 1000*(t1-t0)/i))
      break;
    }
  }
  
  out = mse_stats(theta, mse, Data, verbose=verbose)  
  return(out)
}

#' Helper function for SFP method.
#' Estimates the convexity constant, and the Lipschitz constant.
estimate_SFP_constants = function(Data, nsamples=1000, fctr=10) {
  mle_fit = Data$mle_fit
  Y = Data$Y; X = Data$X; mean_fn = Data$mean_fn
  
  theta_hat = coef(mle_fit)
  S = fctr*vcov(mle_fit)
  vec = eigen(Sigma)$vectors
  Lvals = c()
  Dvals = c()
  for(i in 1:nsamples) {
    theta1 = matrix(rmvnorm(1, mean=theta_hat, sigma=S), ncol=1)
    # theta2 = matrix(rmvnorm(1, mean=theta_hat, sigma=S), ncol=1)
    I = sample(1:p, size=1)
    theta2 = theta1 + vec[, I]
    # theta2 = theta1 + eigen(Sigma)$vectors[, 6]
    h1 = -t(X) %*% (Y - mean_fn(X %*% theta1))/n
    h2 = -t(X) %*% (Y - mean_fn(X %*% theta2))/n
    Lvals = c(Lvals, L2(h1-h2)/L2(theta1-theta2) )
    Dvals = c(Dvals, sum( (h1-h2) * (theta1-theta2)) / L2(theta1-theta2)^2)
  }
  # 
  lvals = eigen(Sigma)$values
  print(sprintf("min eigenval = %.3f (=delta, in linear model) -- max eigenval = %.3f (=L)", min(lvals), max(lvals)))
  print("Returning estimated.")
  
  df = data.frame(delta=min(Dvals), L=max(Lvals))
  print(df)
  return(df)
}

#' Stochastic fixed-point implementation.
#' @param problem_par The problem constants that are useful to use in SFP.
#'                     Can estimate through estimate_SFP_constants()
sfp = function(Data, problem_par=list(), gamma1=.5, is.implicit=FALSE, verbose=TRUE) {

  theta_star = Data$theta_star
  
  # Estimate problem parameters, delta, L
  # Use those values to initialize the procedure.
  if(length(problem_par) == 0) {
    problem_par = estimate_SFP_constants(Data)
  }
  
  L = problem_par$L; delta = problem_par$delta
  a = .1 + log(L / delta)
  K = as.integer(2*a*(1 + gamma1 * L)^2) + 1
  a1 = 2*a/K
  
  if(verbose) {
    print(sprintf("Problem params: gamma1=%.2f, a1=%.5f, K=%d", gamma1, a1, K))
  }
  
  t0 = proc.time()[3]
  theta = rep(0, p)
  mse = c(L2(theta - theta_star))
  for(i in 1:kN) {
    gamma_i = gamma1
    zeta = theta
    #
    for(k in 1:K) {
      ak = a1
      # Main iterations as in Eq.(19) of the main paper.
      if(is.implicit) {
        zeta = (zeta - ak * gamma_i * H(zeta, Data) + ak * theta) / (1 + ak)
      } else {
        zeta = (1-ak) * zeta - ak * (gamma_i * H(zeta, Data) - theta)
      }
    }
    theta = zeta  # update parameter.
    mse = c(mse, L2(theta -  theta_star))
    t1 = proc.time()[3]
    if(t1 - t0 > BUDGET) {
      # budget is over
      if(verbose) {
        print(sprintf("Stopped at i=%d / %d (K=%d). Time = %.2f.  %.3f msec/iter.", 
                      i, kN, K, t1-t0, 1000*(t1-t0)/ (i*K)))
        print("Last estimate of fixed_point = ")
        print(round(theta, 4))
      }
      break;
    } 
  }
  
  out = mse_stats(theta, mse, Data, verbose=verbose)  
  return(out)
}

#' Main code to run a full experiment, including all methods.
#' @param num_gamma How many values of the learning rate to try.
#' @param max_gamma1 Maximum value of learning rate to try.
run_experiment = function(model="normal", num_gamma=10, max_gamma1=25, nreps=10) {
  
  fname = sprintf("%s_experiment.rda", model)
  gamma1_vals = seq(0.1, max_gamma1, length.out=num_gamma)
  methods = c()
  perf = c()
  perf_type = c()
  gammas = c()
  
  # estimate problem constants for SFP
  sfp_par = estimate_SFP_constants(gen_data(model))
  
  pb = txtProgressBar(style=3)
  nall = length(gamma1_vals) * nreps
  cnt = 0
  
  for(g in gamma1_vals) {
    for(j in 1:nreps) {
      ##
      Data_j = gen_data(model)  # refresh data.
      cnt = cnt + 1
      # SGD
      out = sgd(Data_j, gamma1 = g, verbose = F)
      cnames = names(out)
      perf_type = c(perf_type, cnames)
      perf = c(perf, as.numeric(out))
      methods = c(methods, rep("sgd", length(cnames)))
      # ISGD
      out = isgd(Data_j, gamma1 = g, verbose = F)
      perf_type = c(perf_type, cnames)
      perf = c(perf, as.numeric(out))
      methods = c(methods, rep("isgd", length(cnames)))
      # SFP
      out = sfp(Data_j, problem_par=sfp_par, gamma1 = g, verbose = F)
      perf = c(perf, as.numeric(out))
      perf_type = c(perf_type, cnames)
      methods = c(methods, rep("sfp", length(cnames)))
      
      gammas = c(gammas, rep(g, length(cnames)*3))
      ##
      results = data.frame(gamma1=gammas, methods=methods, valtype=perf_type, val=perf)
      setTxtProgressBar(pb, value=cnt/nall)
    }
    
    results = data.frame(gamma1=gammas, methods=methods, valtype=perf_type, val=perf)
    save(results, file=fname)
    # plot_results()
    print(results)
  }
  ##
  results = data.frame(gamma1=gammas, methods=methods, valtype=perf_type, val=perf)
  save(results, file=fname)
  return(results)
}




