#' Robbins-Monro quantile estimation example.
rm(list=ls())
set.seed(123)

#' Indicator function.
I <- function(theta) {
  as.numeric(rnorm(1) <= theta)
} 
filename = "QuantileRegression.rda"

alpha = 0.999 # Solve for F(theta) = a
theta0 = -10
theta_star = qnorm(alpha)
BUDGET = 1 # Both classical RM or proximal RM will run for 1 second.

#' In both RM functions:
#' @param alpha Ground truth. Need to solve F(x) = alpha where F= normal CDF.
#' @param theta0 Starting point.
#'  @param lar Learning rate (gamma_1 in paper)
#' @param N Number of samples.
RM = function(gamma1) {
  # Implements the classical RM procedure.
  stopifnot(gamma1 > 0)
  t0 = proc.time()[3]
  thetas = c(theta0) # parameter estimates
  
  i = 0
  while(TRUE) {
    # Classical RM iteration
    i = i + 1
    gamma_i = gamma1 / i  # learning  rate
    theta_old = tail(thetas, 1)
    y = I(theta_old)
    theta_new = as.numeric(theta_old - gamma_i * (y - alpha))
    # Update.
    thetas = c(thetas, theta_new)
    t1 = proc.time()[3]
    if(t1 - t0 > BUDGET) {
      break
    }
  }
  
  return(thetas)
}

prox_RM = function(gamma1) {
  # Implements Proximal Robbins-Monro
  thetas = c(theta0)
  a1 = 1/(1+gamma1)^2
  K = 50
  t0 = proc.time()[3]
  
  while(TRUE) {
    gamma_i = gamma1  # learning rate fixed due to stoch fixed point.
    theta_old = tail(thetas, 1)
    x1 = theta_old
    
    # SFP algorithm in Eq. (16)
    xk = x1
    for(k in 1:K) {
      ak = a1
      xk = xk - ak * (gamma_i * (I(xk) - alpha) + xk - x1) 
    }
    
    theta_new = as.numeric(xk)
    # Update.
    thetas = c(thetas, theta_new)
    t1 = proc.time()[3]
    if(t1- t0> BUDGET) {
      break;
    }
  }
  return(thetas)
}


total_experiment <- function(max_gamma1=20, nreps=20) {
  # gammas = c(1, 8, 10, 12, 15, 20, 25, 30)
  gammas = seq(.1, max_gamma1, length.out=10)
  A = matrix(NA, nrow=0, ncol=4)
  colnames(A) = c("method", "gamma1", "theta", "alpha")
  pb = txtProgressBar(style=3)
  cnt = 0
  total = length(gammas) * nreps
  
  # NOTE: This could be parallelized over gammas.
  for(gamma1 in gammas) {
    for(j in 1:nreps) {
      cnt = cnt + 1
    
      # Classical RM
      thetas = RM(gamma1)
      theta_rm = as.numeric(tail(thetas, 1))
      A = rbind(A, c("RM", gamma1, theta_rm, alpha))
      # Prox-RM
      thetas = prox_RM(gamma1)
      theta_proxRM = as.numeric(tail(thetas, 1))
      A = rbind(A, c("ProxRM", gamma1, theta_proxRM, alpha))
      
      setTxtProgressBar(pb, value=cnt/total)
      
    }
    A = as.data.frame(A, stringsAsFactors=FALSE)
    save(A, file=filename)
  }
  A = as.data.frame(A, stringsAsFactors=FALSE)
  save(A, file=filename)
}

#' Loads the result from "total_experiment" and creates 
#' a boxplot for the estimates of classical RM and Prox RM 
#' across various learning rates.
#' 
#' @param only.proxRM If true, it only plots the results for prox RM.
gen_boxplot <- function(only.proxRM=FALSE) {
  library(ggplot2)
  load(filename)
  # A = subset(A, method=="ProxRM")
  A$theta = as.numeric(A$theta)
  A$gamma1 = as.numeric(A$gamma1)
  if(only.proxRM) {
    A = subset(A, method=="ProxRM")
  }
  el0 = element_text(size=20)
  el = element_text(size=20)
  el2 = element_text(size=24)
  el3 = element_text(size=24)
  
  library(plyr)
  library(dplyr)
  mean_A = ddply(A, .(method, gamma1), summarize, mean=mean(theta))
  # A = subset(A, method=="ProxRM")
  # A$theta  = sqrt(5 * 1e3) * (A$theta - qnorm(0.999))
  g = ggplot(data = mean_A, aes(x=as.factor(gamma1), y=mean))
  if(only.proxRM) {
    g = ggplot(data = A, aes(x=as.factor(gamma1), y=theta))
    g =  g  + geom_boxplot()
    g = g  + ggtitle("Proximal RM (zoomed-in)")
    g = g + geom_hline(aes(yintercept=theta_star), lty="dashed", lwd=1.25, col="red")
    
  } else  {
    g = g+ ggtitle("Proximal RM (circle) and RM (triangle)")
    g = g + geom_hline(aes(yintercept=theta_star), lty="dashed", lwd=1.25, col="red")
    #  g = g +  geom_boxplot(aes(fill=method, col=method))
    #  g = g + geom_line()
    g = g + geom_point(aes(shape=method, col=method), size=8)
  }
  
  g = g + scale_x_discrete(labels=round(unique(A$gamma1), 2))
  
  g = g + theme(legend.text = element_text(size = rel(1.5)))+
    theme(legend.title = element_text(size = rel(1.5))) + 
    xlab(expression(paste("Learning rate ", gamma[1]))) + 
    ylab(expression(paste("last iterate ", theta["n"])))
  g  = g +  theme(title=el0,
                legend.text=el2,
                legend.title = el2,
                axis.title.x=el,
                axis.text.x =el,
                axis.text.y = el,
                axis.title.y = el)
  
  g = g+theme(legend.position = "none")
  plot(g)
  return(g)
}

#' Reproduces Figure 3 in the main paper.
#' 
gen_Fig3 = function() {
  g1 = gen_boxplot(FALSE)
  g2 = gen_boxplot(TRUE)
  multiplot(g1, g2, cols = 2)
}

multiplot <- function(..., plotlist=NULL, file, cols=1, layout=NULL) {
  library(grid)
  
  # Make a list from the ... arguments and plotlist
  plots <- c(list(...), plotlist)
  
  numPlots = length(plots)
  
  # If layout is NULL, then use 'cols' to determine layout
  if (is.null(layout)) {
    # Make the panel
    # ncol: Number of columns of plots
    # nrow: Number of rows needed, calculated from # of cols
    layout <- matrix(seq(1, cols * ceiling(numPlots/cols)),
                     ncol = cols, nrow = ceiling(numPlots/cols))
  }
  
  if (numPlots==1) {
    print(plots[[1]])
    
  } else {
    # Set up the page
    grid.newpage()
    pushViewport(viewport(layout = grid.layout(nrow(layout), ncol(layout))))
    
    # Make each plot, in the correct location
    for (i in 1:numPlots) {
      # Get the i,j matrix positions of the regions that contain this subplot
      matchidx <- as.data.frame(which(layout == i, arr.ind = TRUE))
      
      print(plots[[i]], vp = viewport(layout.pos.row = matchidx$row,
                                      layout.pos.col = matchidx$col))
    }
  }
}

